/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef SENDS_SSL_BASE_HH
#define SENDS_SSL_BASE_HH
#include "socket_api.hh"
#include <string>

struct ssl_ctx_st;
struct bio_st;


namespace sends {

    /**  OpenSSL context wrapper.
      *  @brief OpenSSL context wrapper.
      *  @author John Zweizig
      *  @version 1.0; Last modified January 23, 2008
      *  @ingroup sends_base
      */
    class ssl_ctx {
    public:
	/**  Construct a new OpenSSL context.
	  *  @brief Constructor.
	  */
	ssl_ctx(void);

	/**  Release and free an OpenSSL context.
	  *  @brief Destructor.
	  */
	virtual ~ssl_ctx(void);

	/**  Throw an exception with the specified text and the OpenSSL
	  *  error text for the most recent error.
	  *  @brief Throw an SSL error exception.
	  *  @exception runtime_error with specified text.
	  *  @param err Caller specified message text.
	  */
	static void error(const std::string& err);

	/**  Secure the context with the specified certificate and key.
	  *  @brief Secure context.
	  *  @param certpath Path to certificate file.
	  *  @param keypath  Path to private key file.
	  *  @param pass     Password to decrypt private key.
	  */
	void secure(const std::string& certpath, const std::string& keypath, 
		    const std::string& pass);

	/**  Construct a new OpenSSL Bio structure.
	  *  @brief Construct a bio.
	  *  @return Pointer to the new Boi structure.
	  */
	bio_st* new_ssl(void);

	/**  Return a reference to the wrapped context structure.
	  *  @brief Reference the context.
	  *  @return Reference to the OpenSSL context structure.
	  */
	ssl_ctx_st& ref_ctx(void);
    private:
	ssl_ctx_st* mCtx;
    };

    //==================================  Ctx inline methods
    inline ssl_ctx_st&
    ssl_ctx::ref_ctx(void) {
	return *mCtx;
    }

    /**  SSL BIO wrapper. Performs I/O and control functions to an SSL port.
      *  Note that this class contains a pointer to a BIO that it owns. When 
      *  the class is copied bu an operator= or a copy constructor, the 
      *  ownership is passed to the new class and the BIO pointer is reset.
      *  This might appear to be an obvious application for the auto_ptr
      *  template, but the BIO must be release by BIO_free rather than 
      *  being deleted.
      *  @brief OpenSSL BIO wrapper.
      *  @author John Zweizig
      *  @version 1.0; Last modified January 23, 2008
      *  @ingroup sends_base
      */
    class ssl_bio :public socket_api {
    public:
	/**  Construct a null SSL BIO wrapper.
	  *  \brief ssl_bio empty constructor.
	  */
	ssl_bio(void);

	/**  Construct an SSL BIO wrapper. The OpenSSL bio is allocated
	  *  from the specified context.
	  *  @brief New ssl_bio constructor.
	  *  @param ctx OpenSSL contect used to construct the BIO structure.
	  */
	explicit ssl_bio(ssl_ctx& ctx);

	/**  Destroy and free the bio.
	  *  @brief Destructor.
	  */
	virtual ~ssl_bio(void);

	/**  Parasitic assignment operator. The Bio of the argument is 
	  *  appropriated by the lhs ssl_bio object.
	  *  @brief Assignment operator.
	  *  @param bio ssl_bio object which is copied.
	  *  @return Reference to this object.
	  */
	ssl_bio& operator=(ssl_bio& bio);

	/**  Accept a connection from a client. THe current ssl_bio should
	  *  be listening to a port (see listen method).
	  *  @brief Accept a client connection.
	  *  @return Pointer to Bio for accepted connetion.
	  */
	ssl_bio* accept(void);

	/**  Complete the handshake on an accepted lient connection.
	  *  @brief Complete client handshake.
	  *  @param port Prt number string.
	  *  @return Socket error code.
	  */
	sock_err bind(const std::string& port);

	/**  Shut-down and free current BIO and zero the internal pointer.
	  *  @brief Close connection.
	  *  @return Socket error code.
	  */
	sock_err close(void);

	/**  Shut-down and free current BIO and zero the internal pointer.
	  *  @brief Close connection.
	  *  @param peer IP address to which the connection will be made.
	  *  @return Socket error code.
	  */
	sock_err connect(const std::string& peer);

	/**  Complete the handshake on an accepted lient connection.
	  *  @brief Complete client handshake.
	  *  @return Socket error code.
	  */
	sock_err handshake(void);

	/**  Listen to the specified port for client connection requests.
	  *  The port to be listened to is identified by a character string
	  *  containing the port number.
	  *  @brief Listen for clients.
	  *  @param port Port ID string.
	  *  @param lq length of request queue.
	  *  @return Socket error code.
	  */
	sock_err listen(const std::string& port, int lq);

	/**  Read a data string via the Bio.
	  *  @brief Read a string
	  *  @param buffer Character vector to receive the data.
	  *  @param length Maximum length of data to be read.
	  *  @return Number of bytes read. 
	  */
	long read_available(char* buffer, long length);

	/**  Reference the wrapped OpenSSL BIO structure.
	  *  @return Modifiable reference to the OpenSSL BIO structure.
	  */
	bio_st& ref_bio(void);

	/**  Release the wrapped OpenSSL BIO structure from this object.
	  *  The Bio structure pointer is is returned and the internal
	  *  pointer is set to null.
	  *  @return Modifiable reference to the OpenSSL BIO structure.
	  */
	bio_st* release(void);

	/**  Free the wrapped bio structure if one exists and reset the 
	  *  internal pointer to the argument value.
	  *  @brief Reset the inter BIO pointer.
	  *  @param bio New value of the bio pointer.
	  */
	void reset(bio_st* bio);

	/**  Set the OpenSSL mode flags in the wrapped BIO.
	  *  @brief Set the SSL mode flags.
	  *  @param mode New mode flags
	  */
	void set_ssl_mode(long mode);

	/**  Get a socket id number. In general this will be the system fd
	  *  for the socket.
	  *  \brief Get the socket id.
	  *  \return Socket id for use in select.
	  */
	int socket_id(void) const;

	/**  Write data string to the socket.
	  *  @brief Write data
	  *  @param buffer Character vector with data to be written.
	  *  @param length Length of data to be written.
	  *  @return Number of bytes written. 
	  */
	long write(const char* buffer, long length);

    private:
	/**  Construct an SSL BIO wrapper. A pointer to the ssl bio is 
	  *  contained in the wrapper class.
	  *  @brief ssl_bio constructor.
	  *  @param bio Pointer to an OpenSSL BIO structure.
	  */
	ssl_bio(bio_st* bio);

    private:
	bio_st* mBio;
    };

    //==================================  Bio inline methods
    inline bio_st&
    ssl_bio::ref_bio(void) {
	return *mBio;
    }

}

#endif //  !defined(SENDS_SSL_BASE_HH)
